from threedi_schema.domain.models import DECLARED_MODELS

from threedi_modelchecker.config import Config

checks = Config(models=DECLARED_MODELS).checks


def test_check_descriptions():
    for check in checks:
        check_code = str(check.error_code).zfill(4)
        try:
            check.level.name.capitalize()
        except Exception as e:
            raise AssertionError(
                f"Could not capitalise level name for check {check_code}"
            ) from e
        try:
            check.description()
        except Exception as e:
            raise AssertionError(
                f"Could not generate description for check {check_code}"
            ) from e



from dataclasses import dataclass
from typing import Optional

from threedi_schema import custom_types

from .base import (
    EnumCheck,
    EPSGGeomCheck,
    ForeignKeyCheck,
    GeometryCheck,
    GeometryTypeCheck,
    NotNullCheck,
    TypeCheck,
    UniqueCheck,
)
from .raster import RasterHasMatchingEPSGCheck


@dataclass
class ForeignKeyCheckSetting:
    col: str
    ref: str
    filter: Optional[bool] = None


def get_level(table, column, level_map):
    level = level_map.get(f"*.{column.name}")
    level = level_map.get(f"{table.name}.{column.name}", level)
    return level or "ERROR"


def generate_foreign_key_checks(table, fk_settings, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    foreign_key_checks = []
    for fk_setting in fk_settings:
        if fk_setting.col.table != table:
            continue
        level = get_level(table, fk_setting.col, custom_level_map)
        # Prevent clash when kwargs contains 'filter'
        filter_val = (
            kwargs.get("filter") if fk_setting.filter is None else fk_setting.filter
        )
        kwargs.pop("filter", None)
        foreign_key_checks.append(
            ForeignKeyCheck(
                reference_column=fk_setting.ref,
                column=fk_setting.col,
                level=level,
                filters=filter_val,
                **kwargs,
            )
        )
    return foreign_key_checks


def generate_unique_checks(
    table, custom_level_map=None, extra_unique_columns=None, **kwargs
):
    custom_level_map = custom_level_map or {}
    unique_checks = []
    if extra_unique_columns is None:
        extra_unique_columns = []
    for column in table.columns:
        if (
            column.unique
            or column.primary_key
            or any(col.compare(column) for col in extra_unique_columns)
        ):
            level = get_level(table, column, custom_level_map)
            unique_checks.append(UniqueCheck(column, level=level, **kwargs))
    return unique_checks


def generate_not_null_checks(
    table, custom_level_map=None, extra_not_null_columns=None, **kwargs
):
    custom_level_map = custom_level_map or {}
    not_null_checks = []
    if extra_not_null_columns is None:
        extra_not_null_columns = []
    for column in table.columns:
        if not column.nullable or any(
            col.compare(column) for col in extra_not_null_columns
        ):
            level = get_level(table, column, custom_level_map)
            not_null_checks.append(NotNullCheck(column, level=level, **kwargs))
    return not_null_checks


def generate_type_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    data_type_checks = []
    for column in table.columns:
        level = get_level(table, column, custom_level_map)
        data_type_checks.append(TypeCheck(column, level=level, **kwargs))
    return data_type_checks


def generate_geometry_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    geometry_checks = []
    for column in table.columns:
        if isinstance(column.type, custom_types.Geometry):
            level = get_level(table, column, custom_level_map)
            geometry_checks.append(GeometryCheck(column, level=level, **kwargs))
    return geometry_checks


def generate_geometry_type_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    geometry_type_checks = []
    for column in table.columns:
        if isinstance(column.type, custom_types.Geometry):
            level = get_level(table, column, custom_level_map)
            geometry_type_checks.append(
                GeometryTypeCheck(column, level=level, **kwargs)
            )
    return geometry_type_checks


def generate_enum_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    enum_checks = []
    for column in table.columns:
        if issubclass(type(column.type), custom_types.CustomEnum):
            level = get_level(table, column, custom_level_map)
            enum_checks.append(EnumCheck(column, level=level, **kwargs))
    return enum_checks


def generate_epsg_geom_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    column = table.columns.get("geom")
    if column is not None:
        level = get_level(table, column, custom_level_map)
        return [EPSGGeomCheck(column=column, level=level, **kwargs)]
    else:
        return []


def generate_epsg_raster_checks(table, raster_columns, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    checks = []
    for column in table.columns:
        if column in raster_columns:
            level = get_level(table, column, custom_level_map)
            checks.append(
                RasterHasMatchingEPSGCheck(column=column, level=level, **kwargs)
            )
    return checks



from unittest import mock

import pytest
from sqlalchemy import select, text
from threedi_schema import constants, models, ThreediDatabase
from threedi_schema.beta_features import BETA_COLUMNS, BETA_VALUES

from threedi_modelchecker.checks.other import (
    AllPresentVegetationParameters,
    BetaColumnsCheck,
    BetaValuesCheck,
    ChannelManholeLevelCheck,
    ConnectionNodesDistance,
    ConnectionNodesLength,
    ControlHasSingleMeasureVariable,
    CorrectAggregationSettingsExist,
    CrossSectionSameConfigurationCheck,
    DefinedAreaCheck,
    DWFDistributionCSVFormatCheck,
    DWFDistributionLengthCheck,
    DWFDistributionSumCheck,
    FeatureClosedCrossSectionCheck,
    InflowNoFeaturesCheck,
    MaxOneRecordCheck,
    ModelEPSGCheckProjected,
    ModelEPSGCheckUnits,
    ModelEPSGCheckValid,
    NodeSurfaceConnectionsCheck,
    OpenChannelsWithNestedNewton,
    PotentialBreachInterdistanceCheck,
    PotentialBreachStartEndCheck,
    PumpStorageTimestepCheck,
    SpatialIndexCheck,
    SurfaceNodeInflowAreaCheck,
    TableControlActionTableCheckDefault,
    TableControlActionTableCheckDischargeCoefficients,
    TagsValidCheck,
    Use0DFlowCheck,
    UsedSettingsPresentCheck,
    UsedSettingsPresentCheckSingleTable,
)
from threedi_modelchecker.model_checks import ThreediModelChecker

from . import factories

SRID = 28992
POINT1 = "142742 473443"
POINT2 = "142743 473443"  # POINT1 + 1 meter
POINT3 = "142747 473443"  # POINT1 + 5 meter


@pytest.mark.parametrize(
    "aggregation_method,flow_variable,expected_result",
    [
        (
            constants.AggregationMethod.CUMULATIVE,
            constants.FlowVariable.PUMP_DISCHARGE,
            0,
        ),  # entries in aggregation settings, valid
        (
            constants.AggregationMethod.CUMULATIVE,
            constants.FlowVariable.DISCHARGE,
            1,
        ),  # entries not in aggregation settings, invalid
    ],
)
def test_aggregation_settings(
    session, aggregation_method, flow_variable, expected_result
):
    factories.ModelSettingsFactory(id=1)
    factories.AggregationSettingsFactory(
        aggregation_method=constants.AggregationMethod.CUMULATIVE,
        flow_variable=constants.FlowVariable.PUMP_DISCHARGE,
    )
    check = CorrectAggregationSettingsExist(
        aggregation_method=aggregation_method, flow_variable=flow_variable
    )
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


def test_connection_nodes_length(session):
    factories.ModelSettingsFactory()
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.ConnectionNodeFactory(id=2, geom=f"SRID={SRID};POINT({POINT2})")
    factories.WeirFactory(
        connection_node_id_start=1,
        connection_node_id_end=2,
    )
    weir_too_short = factories.WeirFactory(
        connection_node_id_start=1,
        connection_node_id_end=1,
    )

    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_id_start,
        end_node=models.Weir.connection_node_id_end,
        min_distance=0.01,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 1
    assert errors[0].id == weir_too_short.id


def test_connection_nodes_length_missing_start_node(session):
    factories.ModelSettingsFactory()
    factories.WeirFactory(connection_node_id_start=1, connection_node_id_end=2)
    factories.ConnectionNodeFactory(id=2)
    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_id_start,
        end_node=models.Weir.connection_node_id_end,
        min_distance=0.05,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 0


def test_connection_nodes_length_missing_end_node(session):
    if session.bind.name == "postgresql":
        pytest.skip("Postgres only accepts coords in epsg 4326")
    factories.ModelSettingsFactory()
    factories.WeirFactory(connection_node_id_start=1, connection_node_id_end=2)
    factories.ConnectionNodeFactory(id=1)
    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_id_start,
        end_node=models.Weir.connection_node_id_end,
        min_distance=0.05,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 0


def test_open_channels_with_nested_newton(session):
    factories.NumericalSettingsFactory(use_nested_newton=0)
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.ConnectionNodeFactory(id=2, geom=f"SRID={SRID};POINT({POINT2})")
    factories.ChannelFactory(
        id=1,
        connection_node_id_start=1,
        connection_node_id_end=2,
        geom=f"SRID={SRID};LINESTRING({POINT1}, {POINT2})",
    )
    factories.CrossSectionLocationFactory(
        channel_id=1,
        cross_section_shape=constants.CrossSectionShape.TABULATED_TRAPEZIUM,
        cross_section_table="0,1\n1,0",
        geom=f"SRID={SRID};POINT({POINT2})",
    )
    factories.ChannelFactory(
        id=2,
        connection_node_id_start=1,
        connection_node_id_end=2,
        geom=f"SRID={SRID};LINESTRING({POINT1}, {POINT2})",
    )
    factories.CrossSectionLocationFactory(
        channel_id=2,
        geom=f"SRID={SRID};POINT({POINT2})",
        cross_section_shape=constants.CrossSectionShape.EGG,
    )

    check = OpenChannelsWithNestedNewton(column=models.CrossSectionLocation.id)

    errors = check.get_invalid(session)
    assert len(errors) == 2


channel_manhole_level_testdata = [
    ("start", -1, -3, -2, 0),
    ("start", -3, -1, -2, 1),
    ("end", -3, -1, -2, 0),
    ("end", -1, -3, -2, 1),
]


@pytest.mark.parametrize(
    "manhole_location,starting_reference_level,ending_reference_level,manhole_level,errors_number",
    channel_manhole_level_testdata,
)
def test_channel_manhole_level_check(
    session,
    manhole_location,
    starting_reference_level,
    ending_reference_level,
    manhole_level,
    errors_number,
):
    # using factories, create one minimal test case which passes, and one which fails
    # once that works, parametrise.
    # use nested factories for channel and connectionNode
    factories.ConnectionNodeFactory(
        id=1,
        geom=f"SRID={SRID};POINT({POINT1})",
        bottom_level=manhole_level,
    )
    factories.ConnectionNodeFactory(
        id=2,
        geom=f"SRID={SRID};POINT({POINT2})",
        bottom_level=manhole_level,
    )
    factories.ChannelFactory(
        id=1,
        geom=f"SRID={SRID};LINESTRING({POINT1}, {POINT2})",
        connection_node_id_start=1,
        connection_node_id_end=2,
    )
    # starting cross-section location
    factories.CrossSectionLocationFactory(
        geom=f"SRID={SRID};POINT(142742.25 473443)",
        reference_level=starting_reference_level,
        channel_id=1,
    )
    # ending cross-section location
    factories.CrossSectionLocationFactory(
        geom=f"SRID={SRID};POINT(142743.25 473443)",
        reference_level=ending_reference_level,
        channel_id=1,
    )
    check = ChannelManholeLevelCheck(nodes_to_check=manhole_location)
    errors = check.get_invalid(session)
    assert len(errors) == errors_number


@pytest.mark.parametrize("other_x, valid", [(142740.05, False), (142740.15, True)])
def test_node_distance_alt(session, other_x, valid):
    factories.ConnectionNodeFactory(id=0, geom=f"SRID={SRID};POINT(142740 473443)")
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({other_x} 473443)")
    # Note that the test uses plain sqlite, so this needs to be committed
    session.commit()
    check = ConnectionNodesDistance(minimum_distance=0.1)
    invalid = check.get_invalid(session)
    assert (len(invalid) == 0) == valid
    # Remove connection nodes
    session.query(models.ConnectionNode).delete()
    session.commit()


class TestCrossSectionSameConfiguration:
    @pytest.mark.parametrize("sep, expected", [(",", "1"), ("\n", "1,2")])
    def test_get_first_in_str(self, session, sep, expected):
        factories.CrossSectionLocationFactory(
            cross_section_table="1,2\n3,4",
        )
        check = CrossSectionSameConfigurationCheck(models.CrossSectionLocation.id)
        result = check.get_first_in_str(
            models.CrossSectionLocation.cross_section_table, sep=sep
        )
        assert session.execute(select(result)).fetchall()[0][0] == expected

    @pytest.mark.parametrize("sep, expected", [(",", "4"), ("\n", "3,4")])
    def test_get_last_in_str(self, session, sep, expected):
        factories.CrossSectionLocationFactory(
            cross_section_table="1,2\n3,4",
        )
        check = CrossSectionSameConfigurationCheck(models.CrossSectionLocation.id)
        result = check.get_last_in_str(
            models.CrossSectionLocation.cross_section_table, sep=sep
        )
        assert session.execute(select(result)).fetchall()[0][0] == expected

    @pytest.mark.parametrize(
        "method_name, expected",
        [
            ("first_row_width", 1.0),
            ("first_row_height", 2.0),
            ("last_row_width", 3.0),
            ("last_row_height", 4.0),
        ],
    )
    def test_row_values(self, session, method_name, expected):
        factories.CrossSectionLocationFactory(
            cross_section_table="1,2\n3,4",
        )
        check = CrossSectionSameConfigurationCheck(models.CrossSectionLocation.id)
        result = getattr(check, method_name)()
        assert session.execute(select(result)).fetchall()[0][0] == expected

    @pytest.mark.parametrize(
        "shape, table, expected",
        [
            (0, "3,4\n3,4", "closed"),
            (1, "3,4\n3,4", "open"),
            (2, "3,4\n3,4", "closed"),
            (3, "3,4\n3,4", "closed"),
            (4, "3,4\n3,4", "open"),
            (5, "3,4\n0,1", "closed"),
            (6, "3,4\n3,4", "open"),
            (7, "3,4\n3,4", "closed"),
            (7, "1,3\n1,4", "open"),
            (8, "3,4\n3,4", "closed"),
        ],
    )
    def test_configuration_type(self, session, shape, table, expected):
        factories.CrossSectionLocationFactory(
            cross_section_shape=shape,
            cross_section_table=table,
        )
        check = CrossSectionSameConfigurationCheck(models.CrossSectionLocation.id)

        cross_sections = (
            select(
                models.CrossSectionLocation.cross_section_shape,
                check.first_row_width().label("first_width"),
                check.first_row_height().label("first_height"),
                check.last_row_width().label("last_width"),
                check.last_row_height().label("last_height"),
            )
            .select_from(models.CrossSectionLocation)
            .subquery()  # Added this line
        )
        # Without this, there is nothing that check.configuration_type can use
        session.execute(select(cross_sections))
        config_types = select(
            check.configuration_type(
                shape=cross_sections.c.cross_section_shape,
                first_width=cross_sections.c.first_width,
                last_width=cross_sections.c.last_width,
                first_height=cross_sections.c.first_height,
                last_height=cross_sections.c.last_height,
            ).label("configuration"),
        )
        assert session.execute(config_types).fetchall()[0][0] == expected


# cases to test: tabulated and non tabulated, everything else is covered above
@pytest.mark.parametrize(
    "shape, width, height, table, same_channels, ok",
    [
        # --- closed cross-sections ---
        # shapes 0, 2, 3 and 8 are always closed
        (0, 3, 4, None, True, False),
        # shape 7 is closed if the first and last (width, height) coordinates are the same
        (
            7,
            None,
            None,
            "2,3\n4.142,0.174\n5.143,0.348\n5.143,0.522\n5.869,0.696\n2,3",
            True,
            False,
        ),
        # shape 7 is open if the first and last (width, height) coordinates are not the same
        (
            7,
            None,
            None,
            "2,4\n4.142,0.174\n5.143,0.348\n5.143,0.522\n5.869,0.696\n3,4",
            True,
            True,
        ),
        # Bad data, could result in false positive but the real issue is covered by other checks
        (7, None, None, "foo", True, False),
        # Check on different channels
        # this should fail if the cross-sections are on the same channel, but pass on different channels
        (0, 3, 4, None, False, True),
    ],
)
def test_cross_section_same_configuration(
    session, shape, width, height, table, same_channels, ok
):
    """
    This test checks two cross-sections on a channel against each other; they should both be open or both be closed.
    In this test, the first cross-section has been set to always be open.
    Therefore, the channel should be invalid when the second cross-section is closed, and valid when it is open.
    """
    factories.ChannelFactory(
        id=1,
    )
    factories.ChannelFactory(
        id=2,
    )
    # shape 1 is always open
    factories.CrossSectionLocationFactory(
        channel_id=1,
        cross_section_shape=1,
        cross_section_width=3,
        cross_section_height=4,
    )
    # the second one is parametrised
    factories.CrossSectionLocationFactory(
        channel_id=1 if same_channels else 2,
        cross_section_shape=shape,
        cross_section_width=width,
        cross_section_height=height,
        cross_section_table=table,
    )
    errors = CrossSectionSameConfigurationCheck(column=models.Channel.id).get_invalid(
        session
    )
    assert len(errors) == (0 if ok else 1)


def test_spatial_index_ok(session):
    check = SpatialIndexCheck(models.ConnectionNode.geom)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


def test_spatial_index_disabled(empty_sqlite_v4):
    session = empty_sqlite_v4.get_session()
    session.execute(text("DROP TABLE IF EXISTS rtree_connection_nodes_geom"))
    check = SpatialIndexCheck(models.ConnectionNode.geom)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


@pytest.mark.parametrize(
    "start,ok",
    [
        ("142742 473443", True),  # at start
        ("142747 473448", True),  # at end
        ("142744.5 473445.5", True),  # middle
        ("142742 473443.01", False),  # close to start
        ("142747 473447.99", False),  # close to end
    ],
)
def test_potential_breach_start_end(session, start, ok):
    # channel geom: "SRID={SRID}};LINESTRING (142742 473443, 142747 473448)"
    factories.ChannelFactory(id=1)
    factories.PotentialBreachFactory(
        geom=f"SRID={SRID};LINESTRING({start}, 142750 473450)", channel_id=1
    )
    check = PotentialBreachStartEndCheck(models.PotentialBreach.geom, min_distance=1.0)
    invalid = check.get_invalid(session)
    if ok:
        assert len(invalid) == 0
    else:
        assert len(invalid) == 1


@pytest.mark.parametrize(
    "start,channel_id,ok",
    [
        (POINT1, 1, True),  # exactly on other
        ("142742.5 473443.5", 1, False),  # too close to other
        ("142742.5 473443.5", 2, True),  # too close to other but other channel
        ("142740 473440", 1, True),  # far enough from other
    ],
)
def test_potential_breach_interdistance(session, start, channel_id, ok):
    # channel geom: "SRID=28992;LINESTRING (142742 473443, 142747 473448)"
    factories.ChannelFactory(id=1)
    factories.ChannelFactory(id=2)
    factories.PotentialBreachFactory(
        geom=f"SRID={SRID};LINESTRING({POINT1}, {POINT3})", channel_id=1
    )
    factories.PotentialBreachFactory(
        geom=f"SRID={SRID};LINESTRING({start}, 142737 473438)", channel_id=channel_id
    )
    check = PotentialBreachInterdistanceCheck(
        models.PotentialBreach.geom, min_distance=1.0
    )
    invalid = check.get_invalid(session)
    if ok:
        assert len(invalid) == 0
    else:
        assert len(invalid) == 1


@pytest.mark.parametrize(
    "storage_area,time_step,expected_result,capacity",
    [
        (0.64, 30, 1, 12500),
        (600, 30, 0, 12500),
        (None, 30, 0, 12500),  # no storage --> open water --> no check
        (600, 30, 0, 0),
    ],
)
def test_pumpstation_storage_timestep(
    session, storage_area, time_step, expected_result, capacity
):
    factories.ConnectionNodeFactory(storage_area=storage_area, id=1)
    factories.PumpFactory(
        connection_node_id=1,
        start_level=-4,
        lower_stop_level=-4.78,
        capacity=capacity,
    )
    factories.TimeStepSettingsFactory(time_step=time_step)
    check = PumpStorageTimestepCheck(models.Pump.capacity)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (1000, 0),  # total area = 1000 + 9000 = 10000 <= 10000; no error
        (1001, 1),  # total area = 1001 + 9000 = 10001 > 10000; error
    ],
)
def test_surface_connection_node_inflow_area(session, value, expected_result):
    connection_node = factories.ConnectionNodeFactory(id=1)
    first_surface = factories.SurfaceFactory(id=1, area=9000)
    second_surface = factories.SurfaceFactory(id=2, area=value)
    factories.SurfaceMapFactory(
        surface_id=first_surface.id, connection_node_id=connection_node.id
    )
    factories.SurfaceMapFactory(
        surface_id=second_surface.id, connection_node_id=connection_node.id
    )
    check = SurfaceNodeInflowAreaCheck()
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "surface_number,expected_result",
    [
        (0, 1),
        (1, 0),
        (10, 0),
    ],
)
def test_inflow_no_features_impervious(session, surface_number, expected_result):
    # add fields
    factories.ModelSettingsFactory()
    if surface_number > 0:
        factories.SurfaceFactory.create_batch(size=surface_number)

    # Only test this for surface because InflowNoFeaturesCheck only uses table length and not table contents
    check = InflowNoFeaturesCheck(feature_table=models.Surface)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "connected_surfaces_count,expected_result",
    [
        (50, 0),
        (51, 1),
    ],
)
def test_connection_node_mapped_surfaces(
    session, connected_surfaces_count, expected_result
):
    factories.ConnectionNodeFactory(id=1)
    for i in range(connected_surfaces_count):
        factories.SurfaceMapFactory(connection_node_id=1, surface_id=i + 1)
    check = NodeSurfaceConnectionsCheck()
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "shape,expected_result",
    [
        (constants.CrossSectionShape.CLOSED_RECTANGLE, 0),
        (constants.CrossSectionShape.RECTANGLE, 1),
    ],
)
def test_feature_closed_cross_section(session, shape, expected_result):
    factories.CrossSectionLocationFactory(
        cross_section_shape=shape, cross_section_width=1, cross_section_height=1
    )
    check = FeatureClosedCrossSectionCheck(models.CrossSectionLocation.id)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "defined_area, max_difference, expected_result",
    [
        (1.2, 0.5, 0),
        (1, 1, 0),
        (2.1, 1, 1),
        (1.6, 0.5, 1),
    ],
)
def test_defined_area(session, defined_area, max_difference, expected_result):
    # create square polygon with area 1
    x0 = int(POINT1.split()[0])
    y0 = int(POINT1.split()[1])
    geom = f"SRID={SRID};POLYGON(({x0} {y0}, {x0 + 1} {y0}, {x0 + 1} {y0 + 1}, {x0} {y0 + 1}, {x0} {y0}))"

    factories.SurfaceFactory(area=defined_area, geom=geom)
    check = DefinedAreaCheck(models.Surface.area, max_difference=max_difference)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (None, 0),  # column not set, valid result
        (True, 1),  # column set, invalid result
    ],
)
def test_beta_columns(session, value, expected_result):
    # Note that the BetaColumnsCheck is just a check for value=None.
    # So it can be mocked with any nullable column
    factories.ModelSettingsFactory(friction_averaging=value)
    check = BetaColumnsCheck(models.ModelSettings.friction_averaging)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (
            constants.BoundaryType.RIEMANN,
            0,
        ),  # column not in beta columns, valid result
        (
            constants.BoundaryType.GROUNDWATERDISCHARGE,
            1,
        ),  # column in beta columns, invalid result
    ],
)
def test_beta_values(session, value, expected_result):
    beta_values = [
        constants.BoundaryType.GROUNDWATERLEVEL,
        constants.BoundaryType.GROUNDWATERDISCHARGE,
    ]
    factories.BoundaryConditions1DFactory(type=value)
    check = BetaValuesCheck(column=models.BoundaryCondition1D.type, values=beta_values)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.skipif(
    condition=(not BETA_COLUMNS and not BETA_VALUES),
    reason="requires beta features to be defined in threedi-schema to run",
)
@pytest.mark.parametrize(
    "allow_beta_features, no_checks_expected",
    [
        (False, False),
        (True, True),
    ],
)
def test_beta_features_in_server(threedi_db, allow_beta_features, no_checks_expected):
    with mock.patch.object(ThreediDatabase, "schema"):
        model_checker = ThreediModelChecker(
            threedi_db, allow_beta_features=allow_beta_features
        )
    model_beta_checks = [
        check
        for check in model_checker.config.checks
        if type(check) in [BetaColumnsCheck, BetaValuesCheck]
    ]
    if no_checks_expected:
        assert len(model_beta_checks) == 0
    else:
        assert len(model_beta_checks) > 0


@pytest.mark.parametrize(
    "cols, shape, friction_type, result",
    [
        # single column defined: should fail
        (
            ["vegetation_height"],
            constants.CrossSectionShape.TABULATED_YZ,
            constants.FrictionType.CHEZY,
            False,
        ),
        # both columns defined, but one empty: should fail
        (
            ["vegetation_height", "vegetation_stem_diameter"],
            constants.CrossSectionShape.TABULATED_YZ,
            constants.FrictionType.CHEZY,
            False,
        ),
        # no columns defined: should pass
        (
            [],
            constants.CrossSectionShape.TABULATED_YZ,
            constants.FrictionType.CHEZY,
            True,
        ),
        # both columns defined: should pass
        (
            [
                "vegetation_drag_coefficient",
                "vegetation_height",
                "vegetation_stem_diameter",
                "vegetation_stem_density",
            ],
            constants.CrossSectionShape.TABULATED_YZ,
            constants.FrictionType.CHEZY,
            True,
        ),
        # shape is not included in check: should pass
        (
            ["vegetation_height"],
            constants.CrossSectionShape.RECTANGLE,
            constants.FrictionType.CHEZY,
            True,
        ),
        # friction type in not included in check: should pass
        (
            ["vegetation_height"],
            constants.CrossSectionShape.TABULATED_YZ,
            constants.FrictionType.MANNING,
            True,
        ),
    ],
)
def test_all_present_fixed_vegetation_parameters(
    session, cols, shape, friction_type, result
):
    veg_args = {col: 1 for col in cols}
    factories.CrossSectionLocationFactory(
        cross_section_shape=shape,
        cross_section_friction_values="1",
        friction_type=friction_type,
        **veg_args,
    )
    check = AllPresentVegetationParameters(
        column=models.CrossSectionLocation.vegetation_height
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == result


@pytest.mark.parametrize(
    "use_0d_inflow",
    [
        constants.InflowType.NO_INFLOW,
        constants.InflowType.SURFACE,
        constants.InflowType.IMPERVIOUS_SURFACE,
    ],
)
@pytest.mark.parametrize("add_surface", [True, False])
def test_use_0d_flow_check(session, use_0d_inflow: int, add_surface: bool):
    factories.SimulationTemplateSettingsFactory(use_0d_inflow=use_0d_inflow)
    if add_surface:
        factories.SurfaceFactory()
    if use_0d_inflow == constants.InflowType.NO_INFLOW:
        nof_invalid_expected = 0
    else:
        nof_invalid_expected = 0 if add_surface else 1
    check = Use0DFlowCheck()
    assert (len(check.get_invalid(session))) == nof_invalid_expected


@pytest.mark.parametrize("use_setting", [True, False])
@pytest.mark.parametrize("add_setting", [True, False])
def test_used_settings_present_check_single_table(session, use_setting, add_setting):
    nof_invalid_expected = 1 if use_setting and not add_setting else 0
    factories.ModelSettingsFactory(use_vegetation_drag_2d=use_setting)
    if add_setting:
        factories.VegetationDragFactory()
    check = UsedSettingsPresentCheckSingleTable(
        column=models.ModelSettings.use_vegetation_drag_2d,
        settings_table=models.VegetationDrag2D,
    )
    assert len(check.get_invalid(session)) == nof_invalid_expected


@pytest.mark.parametrize("use_setting", [True, False])
@pytest.mark.parametrize("add_surface", [True, False])
@pytest.mark.parametrize("add_dwf", [True, False])
def test_used_settings_present_check(session, use_setting, add_surface, add_dwf):
    nof_invalid_expected = 1 if use_setting and not (add_surface and add_dwf) else 0
    factories.SimulationTemplateSettingsFactory(use_0d_inflow=use_setting)
    if add_surface:
        factories.SurfaceFactory()
    if add_dwf:
        factories.DryWeatherFlowFactory()
    check = UsedSettingsPresentCheck(
        column=models.SimulationTemplateSettings.use_0d_inflow,
        settings_tables=[models.Surface, models.DryWeatherFlow],
    )
    assert len(check.get_invalid(session)) == nof_invalid_expected


@pytest.mark.parametrize(
    "nof_rows_to_add, fail",
    [
        (1, False),
        # add to many rows
        (2, True),
        # empty table
        (0, False),
    ],
)
def test_max_one_record_check(session, nof_rows_to_add: int, fail: bool):
    for _ in range(nof_rows_to_add):
        factories.ModelSettingsFactory()
    check = MaxOneRecordCheck(column=models.ModelSettings.id)
    nof_invalid = len(check.get_invalid(session))
    assert (nof_invalid > 0) == fail


def test_tags_valid(session):
    factories.TagsFactory(id=1, description="foo")
    factories.DryWeatherFlowFactory(tags="1,2")
    check = TagsValidCheck(column=models.DryWeatherFlow.tags)
    assert len(check.get_invalid(session)) == 1
    factories.TagsFactory(id=2, description="bar")
    assert len(check.get_invalid(session)) == 0


@pytest.mark.parametrize(
    "action_table, valid",
    [
        ("1,2", True),
        ("1,2\n3,4", True),
        ("2,3\n3,4\n", True),
        ("1.0,2", True),
        ("1,2.1", True),
        ("1;2", False),
        ("1,2 3", False),
        ("1,2,3", False),
    ],
)
def test_control_table_action_table_check_default(session, action_table, valid):
    factories.TableControlFactory(
        action_table=action_table,
        action_type=constants.TableControlActionTypes.set_capacity,
    )
    check = TableControlActionTableCheckDefault()
    assert (len(check.get_invalid(session)) == 0) == valid


@pytest.mark.parametrize(
    "action_table, valid",
    [
        ("1,2 3", True),
        ("1,2 3\n3,4 5", True),
        ("2,3 3\n3,4 5\n", True),
        ("1.0,2 3", True),
        ("1,2.1 3", True),
        ("1,2.1 3", True),
        ("1,2.1 3.3", True),
        ("1,2", False),
        ("1;2 3", False),
        ("1,2,3", False),
        ("1,2 3 4", False),
    ],
)
def test_control_table_action_table_check_discharge_coefficients(
    session, action_table, valid
):
    factories.TableControlFactory(
        action_table=action_table,
        action_type=constants.TableControlActionTypes.set_discharge_coefficients,
    )
    check = TableControlActionTableCheckDischargeCoefficients()
    assert (len(check.get_invalid(session)) == 0) == valid


@pytest.mark.parametrize(
    "measure_variables, valid",
    [
        (
            [
                constants.MeasureVariables.waterlevel,
                constants.MeasureVariables.waterlevel,
            ],
            True,
        ),
        (
            [constants.MeasureVariables.waterlevel, constants.MeasureVariables.volume],
            False,
        ),
    ],
)
def test_control_has_single_measure_variable(session, measure_variables, valid):
    factories.TableControlFactory(id=1)
    for i, measure_variable in enumerate(measure_variables, 1):
        factories.MeasureMapFactory(
            control_id=1, measure_location_id=i, control_type="table"
        )
        factories.MeasureLocationFactory(id=i, measure_variable=measure_variable)
    check = ControlHasSingleMeasureVariable(control_model=models.TableControl)
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize("distribution, valid", [("1,2", True), ("1 2", False)])
def test_dwf_distribution_csv_format_check(session, distribution, valid):
    factories.DryWeatherFlowDistributionFactory(distribution=distribution)
    check = DWFDistributionCSVFormatCheck()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize(
    "distribution, valid",
    [
        (
            "3,1.5,1,1,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,4",
            True,
        ),
        ("1,2", False),
    ],
)
def test_dwf_distribution_length_check(session, distribution, valid):
    factories.DryWeatherFlowDistributionFactory(distribution=distribution)
    check = DWFDistributionLengthCheck()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize(
    "distribution, valid",
    [
        (
            "3,1.5,1,1,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,4",
            True,
        ),
        (
            "3.33,1.5,1.33,0.33,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,4",
            True,
        ),
        (
            "3.3,1.5,1.3,0.3,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,4",
            False,
        ),
        (
            "3,1.5,1,1,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,40",
            False,
        ),
    ],
)
def test_dwf_distribution_sum_check(session, distribution, valid):
    factories.DryWeatherFlowDistributionFactory(distribution=distribution)
    check = DWFDistributionSumCheck()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize(
    "ref_epsg, valid",
    [
        (28992, True),
        (999999, False),
    ],
)
def test_model_epsg_check_valid(session, ref_epsg, valid):
    factories.ModelSettingsFactory()
    session.model_checker_context.epsg_ref_code = ref_epsg
    check = ModelEPSGCheckValid()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize(
    "ref_epsg, valid",
    [
        (28992, True),
        (4326, False),
        (None, True),  # skip check when there is no epsg
    ],
)
def test_model_epsg_check_projected(session, ref_epsg, valid):
    factories.ModelSettingsFactory()
    session.model_checker_context.epsg_ref_code = ref_epsg
    check = ModelEPSGCheckProjected()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid


@pytest.mark.parametrize(
    "ref_epsg, valid",
    [
        (28992, True),
        (4326, False),
        (None, True),  # skip check when there is no epsg
    ],
)
def test_model_epsg_check_units(session, ref_epsg, valid):
    factories.ModelSettingsFactory()
    session.model_checker_context.epsg_ref_code = ref_epsg
    check = ModelEPSGCheckUnits()
    invalids = check.get_invalid(session)
    assert (len(invalids) == 0) == valid



import csv
from io import StringIO
from typing import NamedTuple

from threedi_modelchecker.checks.base import BaseCheck


# error handling export functions
def print_errors(errors):
    """Simply prints all errors to stdout

    :param errors: iterator of BaseModelError
    """
    for error in errors:
        print(format_check_results(*error))


def export_to_file(errors, file):
    """Write errors to a new file, separated with newlines.

    File cannot be an already existing file.

    :param errors: iterator of BaseModelError
    :param file:
    :return: None
    :raise FileExistsError: if the file already exists
    """
    with open(file, "w") as f:
        for error in errors:
            f.write(format_check_results(*error) + "\n")


def format_check_results(check: BaseCheck, invalid_row: NamedTuple):
    OUTPUT_FORMAT = "{level}{error_code:04d} (id={row_id:d}) {description!s}"
    return OUTPUT_FORMAT.format(
        level=check.level.name[:1],
        error_code=check.error_code,
        row_id=invalid_row.id,
        description=check.description(),
    )


# check overview export functions
def order_checks(checks) -> list:
    """
    Alphabetically order checks so that they will consistently be ordered the same.
    This orders first by error code, and then checks with the same error code are
    sorted alphabetically by description. Checks are not sorted by level, because
    in general, checks with the same error code have the same level.

    This makes Github PR requests a lot clearer.
    """
    return sorted(checks, key=lambda k: (k.error_code, k.description()))


def generate_rst_table(checks) -> str:
    "Generate an RST table to copy into the Sphinx docs with a list of checks"
    rst_table_string = ""
    header = (
        ".. list-table:: Executed checks\n"
        + "   :widths: 10 20 40\n"
        + "   :header-rows: 1\n\n"
        + "   * - Check number\n"
        + "     - Check level\n"
        + "     - Check message"
    )
    rst_table_string += header
    for check in order_checks(checks):
        # pad error code with leading zeroes so it is always 4 numbers
        formatted_error_code = str(check.error_code).zfill(4)
        check_row = (
            "\n"
            + f"   * - {formatted_error_code}\n"
            + f"     - {check.level.name.capitalize()}\n"
            + f"     - {check.description()}"
        )
        rst_table_string += check_row
    return rst_table_string


def generate_csv_table(checks) -> str:
    "Generate an CSV table with a list of checks for use elsewhere"
    # a StringIO buffer is used so that the CSV can be printed to terminal as well as written to file
    output_buffer = StringIO()
    fieldnames = ["error_code", "level", "description"]
    writer = csv.DictWriter(
        output_buffer, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
    )

    writer.writeheader()

    checks = order_checks(checks)

    for check in checks:
        writer.writerow(
            {
                "error_code": check.error_code,
                "level": check.level.name,
                "description": check.description(),
            }
        )

    return output_buffer.getvalue()



import pytest

from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.exporters import generate_csv_table, generate_rst_table


@pytest.fixture
def fake_checks():
    class FakeCheck:
        def __init__(self, level, error_code):
            self.level = level
            self.error_code = error_code

        def description(self):
            return f"This sample message has code {self.error_code} and level {self.level.name}"

    fake_checks = [
        FakeCheck(level=CheckLevel.WARNING, error_code=2),
        FakeCheck(level=CheckLevel.ERROR, error_code=1234),
        FakeCheck(level=CheckLevel.INFO, error_code=12),
    ]

    return fake_checks


def test_generate_rst_table(fake_checks):
    correct_rst_result = (
        ".. list-table:: Executed checks\n"
        + "   :widths: 10 20 40\n   :header-rows: 1\n\n"
        + "   * - Check number\n"
        + "     - Check level\n"
        + "     - Check message\n"
        + "   * - 0002\n"
        + "     - Warning\n"
        + "     - This sample message has code 2 and level WARNING\n"
        + "   * - 0012\n"
        + "     - Info\n"
        + "     - This sample message has code 12 and level INFO\n"
        + "   * - 1234\n"
        + "     - Error\n"
        + "     - This sample message has code 1234 and level ERROR"
    )
    rst_result = generate_rst_table(fake_checks)
    assert rst_result == correct_rst_result


def test_generate_csv_table(fake_checks):
    correct_csv_result = (
        '"error_code","level","description"\r\n'
        + '2,"WARNING","This sample message has code 2 and level WARNING"\r\n'
        + '12,"INFO","This sample message has code 12 and level INFO"\r\n'
        + '1234,"ERROR","This sample message has code 1234 and level ERROR"\r\n'
    )
    csv_result = generate_csv_table(fake_checks)
    assert csv_result == correct_csv_result



from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Union

from sqlalchemy import func
from threedi_schema import constants, models

from .base import BaseCheck


class CrossSectionTableColumnIdx(IntEnum):
    height = 0
    width = 1
    all = 2


class CrossSectionTableXYColumnIdx(IntEnum):
    Y = 0
    Z = 1
    all = 2


def parse_csv_table_col(str_data, idx):
    return [float(line.split(",")[idx]) for line in str_data.splitlines()]


def parse_csv_table(str_data):
    return [[float(item) for item in line.split(",")] for line in str_data.splitlines()]


class CrossSectionBaseCheck(BaseCheck):
    """Base class for all cross section definition checks."""

    def __init__(self, column, *args, **kwargs):
        self.shapes = kwargs.pop("shapes", None)
        super().__init__(column, *args, **kwargs)

    @property
    def shape_msg(self):
        if self.shapes is None:
            return ["all"]
        return sorted({x.value for x in self.shapes})

    def to_check(self, session):
        qs = super().to_check(session)
        if self.shapes is not None:
            qs = qs.filter(self.table.c.cross_section_shape.in_(self.shapes))
        return qs

    def parse_cross_section_vegetation_table(self, session):
        column = models.CrossSectionLocation.cross_section_vegetation_table
        records = self.to_check(session).filter((column != None) & (column != ""))
        for record in records:
            try:
                table = parse_csv_table(getattr(record, column.name))
                if len(table[0]) != 4:
                    raise ValueError(
                        "cross_section_vegetation_table must have 4 columns"
                    )
                yield (record, table)
            except (IndexError, ValueError):
                continue  # Skip records with errors

    def parse_cross_section_table(
        self,
        session,
        col_idx: Union[CrossSectionTableColumnIdx, CrossSectionTableXYColumnIdx],
    ):
        column = self.table.c.cross_section_table
        records = self.to_check(session).filter((column != None) & (column != ""))
        for record in records:
            try:
                if col_idx.name == "all":
                    str = getattr(record, column.name)
                    values = tuple(
                        [
                            parse_csv_table_col(str, _col_idx.value)
                            for _col_idx in type(col_idx)
                            if _col_idx != col_idx
                        ]
                    )
                else:
                    values = parse_csv_table_col(
                        getattr(record, column.name), col_idx.value
                    )
                yield (record, values)  # Yield the parsed values as a generator
            except (IndexError, ValueError):
                continue  # Skip records with errors


class CrossSectionNullCheck(CrossSectionBaseCheck):
    """Check if width / height is not NULL or empty"""

    def get_invalid(self, session):
        return (
            self.to_check(session)
            .filter((self.column == None) | (self.column == ""))
            .all()
        )

    def description(self):
        return f"{self.column_name} cannot be null or empty for shapes {self.shape_msg}"


class CrossSectionExpectEmptyCheck(CrossSectionBaseCheck):
    """Check if width / height is NULL or empty"""

    def get_invalid(self, session):
        return (
            self.to_check(session)
            .filter((self.column != None) & (self.column != ""))
            .all()
        )

    def description(self):
        return f"{self.column_name} should be null or empty for shapes {self.shape_msg}"


class CrossSectionGreaterZeroCheck(CrossSectionBaseCheck):
    """Check that width / height is larger than 0"""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                value = getattr(record, self.column.name)
            except ValueError:
                continue

            if value <= 0:
                invalids.append(record)
        return invalids

    def description(self):
        return f"{self.column_name} should be greater than zero for shapes {self.shape_msg}"


class CrossSectionCSVFormatCheck(CrossSectionBaseCheck):
    """Check whether each row in the string can be converted to a list of floats"""

    ncol = None
    nrow = None

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                lines = getattr(record, self.column.name).splitlines()
                if self.nrow is not None and len(lines) != self.nrow:
                    invalids.append(record)
                    continue
                for line in lines:
                    line = line.split(",")
                    if self.ncol is not None and len(line) != self.ncol:
                        invalids.append(record)
                        break
                    for x in line:
                        float(x)
            except ValueError:
                invalids.append(record)

        return invalids

    def description(self):
        return (
            f"{self.table.name}.{self.column_name} should contain one or more lines of "
            "comma separated values"
        )


class CrossSectionListCheck(CrossSectionCSVFormatCheck):
    nrow = 1

    def description(self):
        return (
            f"{self.table.name}.{self.column_name} should contain comma separated floats "
            f"for shapes {self.shape_msg}"
        )


class CrossSectionTableCheck(CrossSectionCSVFormatCheck):
    """Check whether each row in the table contains a list of ncols comma separated floats"""

    def __init__(self, ncol, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ncol = ncol

    def description(self):
        return (
            f"{self.column_name} should contain a csv table containing "
            f"{self.ncol} columns with floats for shapes {self.shape_msg}"
        )


class CrossSectionIncreasingCheck(CrossSectionBaseCheck):
    """Tabulated definitions should have an increasing list of heights."""

    def get_invalid(self, session):
        invalids = []
        for record, values in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableColumnIdx.height
        ):
            if values != sorted(values):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should be monotonically increasing for shapes {self.shape_msg}. Maybe the width and height have been interchanged?"


class CrossSectionFirstElementZeroCheck(CrossSectionBaseCheck):
    """Tabulated definitions should start at with 0 height."""

    def get_invalid(self, session):
        invalids = []
        for record, values in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableColumnIdx.height
        ):
            if abs(values[0]) != 0:
                invalids.append(record)

        return invalids

    def description(self):
        return f"The first element of {self.column_name} should equal 0 for shapes {self.shape_msg}. Note that heights are relative to 'reference_level'."


class CrossSectionFirstElementNonZeroCheck(CrossSectionBaseCheck):
    """Tabulated rectangles cannot start with 0 width."""

    def get_invalid(self, session):
        invalids = []
        for record, values in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableColumnIdx.width
        ):
            if abs(values[0]) <= 0:
                invalids.append(record)

        return invalids

    def description(self):
        return f"The first element of {self.column_name} must be larger than 0 for tabulated rectangle shapes. Consider using tabulated trapezium."


class CrossSectionYZHeightCheck(CrossSectionBaseCheck):
    """The height of a yz profile should include 0 and should not have negative
    elements.
    """

    def get_invalid(self, session):
        invalids = []
        for record, values in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableXYColumnIdx.Z
        ):
            if any(x < 0 for x in values) or not any(x == 0 for x in values):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} for YZ profiles should include 0.0 and should not include negative values."


class CrossSectionYZCoordinateCountCheck(CrossSectionBaseCheck):
    """yz profiles should have 3 coordinates (excluding a closing coordinate)"""

    def get_invalid(self, session):
        invalids = []
        for record, (Y, Z) in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableXYColumnIdx.all
        ):
            if len(Y) == 0 or len(Y) != len(Z):
                continue
            is_closed = Z[0] == Z[-1] and Y[0] == Y[-1]
            if len(Z) < (4 if is_closed else 3):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.table.name} width and height should contain at least 3 coordinates (excluding closing coordinate) for YZ profiles"


class CrossSectionYZIncreasingWidthIfOpenCheck(CrossSectionBaseCheck):
    """yz profiles should have increasing widths (y) if they are open"""

    def get_invalid(self, session):
        invalids = []
        for record, (Y, Z) in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableXYColumnIdx.all
        ):
            if Y[0] == Y[-1] and Z[0] == Z[-1]:
                continue
            elif len(Y) > 1 and any(
                previous_width >= next_width
                for (previous_width, next_width) in zip(Y[:-1], Y[1:])
            ):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should be strictly increasing for open YZ profiles. Perhaps this is actually a closed profile?"


def get_widths_heights_for_tabulated_record(record):
    if not record.cross_section_shape.is_tabulated:
        raise ValueError(
            "get_widths_heighs_for_tabulated_record cannot handle tabulated shaptes"
        )
    if record.cross_section_shape == constants.CrossSectionShape.TABULATED_YZ:
        widths = parse_csv_table_col(
            record.cross_section_table, CrossSectionTableXYColumnIdx.Y
        )
        heights = parse_csv_table_col(
            record.cross_section_table, CrossSectionTableXYColumnIdx.Z
        )
    else:
        widths = parse_csv_table_col(
            record.cross_section_table, CrossSectionTableColumnIdx.width
        )
        heights = parse_csv_table_col(
            record.cross_section_table, CrossSectionTableColumnIdx.height
        )
    return widths, heights


def cross_section_configuration_for_record(record):
    if record.cross_section_shape.is_tabulated:
        widths, heights = get_widths_heights_for_tabulated_record(record)
        return cross_section_configuration_tabulated(
            shape=record.cross_section_shape, widths=widths, heights=heights
        )
    else:
        return cross_section_configuration_not_tabulated(
            shape=record.cross_section_shape,
            width=record.cross_section_width,
            height=record.cross_section_height,
        )


def cross_section_configuration_not_tabulated(shape, width, height):
    """
    Retrieve maximum width, maximum height  and open/closed configuration for not tabulated
    cross-sections.
    """
    if shape.is_tabulated:
        raise ValueError("cross_section_configuration cannot handle tabulated shaptes")
    max_width = 0 if not width else width
    if shape == constants.CrossSectionShape.CLOSED_RECTANGLE:
        max_height = 0 if not height else height
    elif shape == constants.CrossSectionShape.RECTANGLE:
        max_height = height
    elif shape == constants.CrossSectionShape.CIRCLE:
        max_height = max_width
    elif shape in [
        constants.CrossSectionShape.EGG,
        constants.CrossSectionShape.INVERTED_EGG,
    ]:
        max_height = 1.5 * max_width
    configuration = "closed" if shape.is_closed else "open"
    return max_width, max_height, configuration


def cross_section_configuration_tabulated(shape, widths, heights):
    """
    Retrieve maximum width, maximum height  and open/closed configuration for tabulated cross-sections.
    """
    if not shape.is_tabulated:
        raise ValueError(
            "cross_section_configuration_tabulated can only handle tabulated shaptes"
        )
    if not widths:
        widths = [0]
    if not heights:
        heights = [0]
    if shape in [
        constants.CrossSectionShape.TABULATED_RECTANGLE,
        constants.CrossSectionShape.TABULATED_TRAPEZIUM,
    ]:
        last_width = widths[-1]
        max_height = max(heights)
        max_width = max(widths)
        if last_width == 0:
            configuration = "closed"
        elif last_width > 0:
            configuration = "open"
        else:
            raise ValueError(
                "A tabulated rectangle or trapezium cannot have a negative last width"
            )
    elif shape == constants.CrossSectionShape.TABULATED_YZ:
        # without the rounding, floating-point errors occur
        max_width = round((max(widths) - min(widths)), 9)
        max_height = round((max(heights) - min(heights)), 9)
        first_width = widths[0]
        last_width = widths[-1]
        first_height = heights[0]
        last_height = heights[-1]
        if (first_width, first_height) == (last_width, last_height):
            configuration = "closed"
        else:
            configuration = "open"
    else:
        raise ValueError(
            "cross_section_configuration_tabulated was called for a tabulated shape other "
            "than TABULATED_YZ, TABULATED_RECTANGLE or TABULATED_TRAPEZIUM"
        )
    return max_width, max_height, configuration


class CrossSectionMinimumDiameterCheck(CrossSectionBaseCheck):
    """Check if cross section widths and heights are large enough"""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session):
            (
                max_width,
                max_height,
                configuration,
            ) = cross_section_configuration_for_record(record)
            # See nens/threedi-modelchecker#251
            minimum_diameter = 0.1
            if configuration == "closed":
                if (max_height < minimum_diameter) or (max_width < minimum_diameter):
                    invalids.append(record)
            # the profile height does not need checking on an open cross-section
            elif configuration == "open":
                if max_width < minimum_diameter:
                    invalids.append(record)

        return invalids

    def description(self):
        return f"{self.table.name}.cross_section_width and/or cross_section_height should be at least 0.1m"

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .where(
                self.table.c.cross_section_shape.isnot(None),
            )
        )


class OpenIncreasingCrossSectionCheck(CrossSectionBaseCheck):
    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session):
            # friction with conveyance can only be used for cross-sections
            # which are open *and* have a monotonically increasing width
            if record.cross_section_shape.is_tabulated:
                if (
                    record.cross_section_shape
                    == constants.CrossSectionShape.TABULATED_YZ
                ):
                    widths = parse_csv_table_col(
                        record.cross_section_table, CrossSectionTableXYColumnIdx.Y.value
                    )
                else:
                    widths = parse_csv_table_col(
                        record.cross_section_table, CrossSectionTableColumnIdx.width
                    )
                if any(
                    next_width < previous_width
                    for (previous_width, next_width) in zip(widths[:-1], widths[1:])
                ):
                    invalids.append(record)
                    continue
            _, _, configuration = cross_section_configuration_for_record(record)
            if configuration == "closed":
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} can only be used in an open channel with monotonically increasing width values"

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .where(self.table.c.cross_section_shape.isnot(None))
        )


class OpenIncreasingCrossSectionConveyanceFrictionCheck(
    OpenIncreasingCrossSectionCheck
):
    """
    Check if cross sections used with friction with conveyance
    are open and monotonically increasing in width
    """

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .where(
                self.table.c.cross_section_shape.isnot(None),
                self.table.c.friction_type.in_(
                    [
                        constants.FrictionType.CHEZY_CONVEYANCE,
                        constants.FrictionType.MANNING_CONVEYANCE,
                    ]
                ),
            )
        )

    def description(self):
        return (
            f"{self.table.name}.friction_type can only "
            "have conveyance if the associated definition is "
            "an open shape, and its width is monotonically increasing"
        )


class OpenIncreasingCrossSectionVariableCheck(OpenIncreasingCrossSectionCheck):
    """
    Check if cross sections used with friction with conveyance
    are open and monotonically increasing in width
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            shapes=(constants.CrossSectionShape.TABULATED_YZ,), *args, **kwargs
        )

    def description(self):
        return f"{self.column_name} can only be used in an open channel with monotonically increasing width values"


class CrossSectionVariableCorrectLengthCheck(CrossSectionBaseCheck, ABC):
    """Variable friction and vegetation properties should contain 1 value for each element; len(var_property) = len(width)-1"""

    @abstractmethod
    def parse_str_value(self, variable):
        return

    def get_invalid(self, session):
        invalids = []
        for record, widths in self.parse_cross_section_table(
            session, col_idx=CrossSectionTableColumnIdx.width
        ):
            try:
                values = self.parse_str_value(getattr(record, self.column.name))

            except ValueError:
                continue  # other check catches this
            if not (len(widths) - 1 == len(values)):
                invalids.append(record)
        return invalids

    def description(self):
        return f"{self.column_name} should contain 1 value for each element."


class OpenIncreasingCrossSectionConveyanceFrictionCheckWithMaterial(
    OpenIncreasingCrossSectionCheck
):
    """
    Check if cross sections used with friction with conveyance
    are open and monotonically increasing in width
    """

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .outerjoin(models.Material, self.table.c.material_id == models.Material.id)
            .where(
                self.table.c.cross_section_shape.isnot(None),
                # take value from table if present, otherwise value from Material
                func.coalesce(
                    self.table.c.friction_type, models.Material.friction_type
                ).in_(
                    [
                        constants.FrictionType.CHEZY_CONVEYANCE,
                        constants.FrictionType.MANNING_CONVEYANCE,
                    ]
                ),
            )
        )

    def description(self):
        return f"{self.column_name} should contain 1 value for each element; len({self.column_name}) = len(width)-1"


class CrossSectionFrictionCorrectLengthCheck(CrossSectionVariableCorrectLengthCheck):
    def parse_str_value(self, str_value):
        return [float(item) for item in str_value.split(",")]


class CrossSectionVegetationCorrectLengthCheck(CrossSectionVariableCorrectLengthCheck):
    def parse_str_value(self, str_value):
        return list(str_value.splitlines())


class CrossSectionVegetationTableNotNegativeCheck(CrossSectionBaseCheck):
    """Check for negative valuesin cross_section_vegetation_table"""

    def get_invalid(self, session):
        invalids = []
        for record, table in self.parse_cross_section_vegetation_table(session):
            if any(item < 0 for row in table for item in row):
                invalids.append(record)
        return invalids

    def description(self):
        return f"All values in {self.table.name}.{self.column_name} should be equal to or larger than 0"


class CrossSectionVariableFrictionRangeCheck(CrossSectionBaseCheck):
    def __init__(
        self,
        friction_types,
        min_value=None,
        max_value=None,
        left_inclusive=True,
        right_inclusive=True,
        message=None,
        *args,
        **kwargs,
    ):
        self.friction_types = friction_types
        if min_value is None and max_value is None:
            raise ValueError("Please supply at least one of {min_value, max_value}.")
        str_parts = []
        if min_value is None:
            self.min_valid = lambda x: True
        else:
            self.min_valid = (
                (lambda x: x >= min_value)
                if left_inclusive
                else (lambda x: x > min_value)
            )
            str_parts.append(f"{'< ' if left_inclusive else '<= '}{min_value}")
        if max_value is None:
            self.max_valid = lambda x: True
        else:
            self.max_valid = (
                (lambda x: x <= max_value)
                if right_inclusive
                else (lambda x: x < max_value)
            )
            str_parts.append(f"{'> ' if right_inclusive else '>= '}{max_value}")
        self.range_str = " and/or ".join(str_parts)
        self.message = message
        super().__init__(*args, **kwargs)

    def get_invalid(self, session):
        invalids = []
        records = (
            self.to_check(session)
            .filter((self.column != None) & (self.column != ""))
            .filter(
                models.CrossSectionLocation.friction_type.in_(self.friction_types)
                & models.CrossSectionLocation.cross_section_friction_values.is_not(None)
            )
        )
        for record in records:
            try:
                values = [
                    float(x) for x in getattr(record, self.column.name).split(",")
                ]
            except ValueError:
                invalids.append(record)
            if not self.min_valid(min(values)):
                invalids.append(record)
            elif not self.max_valid(max(values)):
                invalids.append(record)
        return invalids

    def description(self):
        if self.message is None:
            return f"some values in {self.column_name} are {self.range_str}"
        else:
            return self.message



from dataclasses import dataclass
from math import isclose
from pathlib import Path
from typing import Dict, Optional, Type

from threedi_schema import models

from threedi_modelchecker.interfaces import GDALRasterInterface, RasterInterface

from .base import BaseCheck

# interface_cls.is_valid_geotiff returns False if the dataset is empty
# (such as when the column doesn't reference a file) or the dataset is not a valid geotiff.
# It's used on every raster check which reads the raster file in any way,
# to prevent any other code in the check from executing on a bad file.
# On such datasets, RasterExistsCheck and/or RasterIsValidCheck will be raised.


class Context:
    pass


@dataclass
class ServerContext(Context):
    available_rasters: Dict[str, str]
    raster_interface: Type[RasterInterface] = GDALRasterInterface
    epsg_ref_code: int = None
    epsg_ref_name: str = ""


@dataclass
class LocalContext(Context):
    base_path: Path
    raster_interface: Type[RasterInterface] = GDALRasterInterface
    epsg_ref_code: int = None
    epsg_ref_name: str = ""


class BaseRasterCheck(BaseCheck):
    """Baseclass for all raster checks.

    Because these checks are different on local and server systems, subclasses may
    implement 2 methods: is_valid_local and/or is_valid_local.
    """

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .filter(
                self.column != None,
                self.column != "",
            )
        )

    def get_invalid(self, session):
        context = session.model_checker_context
        raster_interface = context.raster_interface
        if not raster_interface.available():
            return []
        if isinstance(context, ServerContext):
            # max 1 record; we can only have 1 raster per type on the server
            records = list(self.to_check(session).limit(1).all())
            paths = [self.get_path_server(x, context) for x in records]
        else:
            records = list(self.to_check(session).all())
            paths = [self.get_path_local(x, context) for x in records]
        return [
            record
            for (record, path) in zip(records, paths)
            if path is not None and not self.is_valid(path, raster_interface)
        ]

    def get_path_local(self, record, context: LocalContext) -> Optional[str]:
        abs_path = context.base_path.joinpath(
            "rasters", getattr(record, self.column.name)
        )
        if abs_path.exists():
            return str(abs_path)

    def get_path_server(self, record, context: ServerContext) -> str:
        if isinstance(context.available_rasters, dict):
            return context.available_rasters.get(self.column.name)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        return True


class RasterExistsCheck(BaseRasterCheck):
    """Check whether a file referenced in given Column exists.

    In order to perform this check, the SQLAlchemy session requires a
    `model_checker_context` attribute, which is set automatically by the
    ThreediModelChecker and contains either a LocalContext or
    ServerContextinstance.
    """

    def get_invalid(self, session):
        context = session.model_checker_context
        if isinstance(context, ServerContext):
            # map names of raster files fields in the schema to those in the RasterOptions class
            raster_name_mapping = {
                "friction_coefficient_file": "frict_coef_file",
                "initial_water_level_file": "initial_waterlevel_file",
                "groundwater_hydraulic_conductivity_file": "groundwater_hydro_connectivity_file",
                "max_infiltration_volume_file": "max_infiltration_capacity_file",
            }
            raster_col = self.column.name
            if self.column.name in raster_name_mapping:
                raster_col = raster_name_mapping[self.column.name]
            if raster_col in context.available_rasters:
                return []
            else:
                return self.to_check(session).all()
        else:
            records = list(self.to_check(session).all())
            paths = [self.get_path_local(x, context) for x in records]
            return [record for (record, path) in zip(records, paths) if path is None]

    def description(self):
        return f"The file in {self.column_name} is not present"


class RasterIsValidCheck(BaseRasterCheck):
    """Check whether a file is a geotiff."""

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            return raster.is_valid_geotiff

    def description(self):
        return f"The file in {self.column_name} is not a valid GeoTIFF file"


class RasterHasOneBandCheck(BaseRasterCheck):
    """Check whether a raster has a single band."""

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            return raster.band_count == 1

    def description(self):
        return f"The file in {self.column_name} has multiple or no bands."


class RasterHasMatchingEPSGCheck(BaseRasterCheck):
    """Check whether a raster's EPSG code matches the EPSG code in the global settings for the SQLite."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.epsg_ref_name = ""
        self.epsg_ref_code = None

    def get_invalid(self, session):
        if session.model_checker_context.epsg_ref_code is None:
            return []
        self.epsg_ref_name = session.model_checker_context.epsg_ref_name
        self.epsg_ref_code = session.model_checker_context.epsg_ref_code
        return super().get_invalid(session)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        if self.epsg_ref_code is None:
            return True
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff or not raster.has_projection:
                return True
            if raster.epsg_code is None:
                return False
            return raster.epsg_code == self.epsg_ref_code

    def description(self):
        return f"The file in {self.column_name} has no EPSG code or the EPSG code does not match does not match {self.epsg_ref_name}"


class RasterSquareCellsCheck(BaseRasterCheck):
    """Check whether a raster has square cells (pixels)"""

    def __init__(self, *args, decimals=7, **kwargs):
        super().__init__(*args, **kwargs)
        self.decimals = decimals

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            dx, dy = raster.pixel_size
            return dx is not None and round(dx, self.decimals) == round(
                dy, self.decimals
            )

    def description(self):
        return f"The raster in {self.column_name} has non-square raster cells."


class RasterGridSizeCheck(BaseRasterCheck):
    """Check whether the global settings' grid size is an even multiple of a raster's cell size (at least 2x)."""

    def get_invalid(self, session):
        minimum_cell_size_query = session.query(
            models.ModelSettings.minimum_cell_size
        ).first()
        if minimum_cell_size_query is not None:
            self.minimum_cell_size = minimum_cell_size_query[0]
        else:
            return []
        return super().get_invalid(session)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            # the x pixel size is used here,but it is equal to the y pixel size
            try:
                return (
                    isclose(
                        a=((self.minimum_cell_size / raster.pixel_size[0]) % 2),
                        b=0,
                        rel_tol=1e-09,
                    )
                ) and (self.minimum_cell_size >= (2 * raster.pixel_size[0]))
            # if one of the fields is a NoneType it will be caught elsewhere
            except TypeError:
                return True

    def description(self):
        return "model_settings.minimum_cell_size is not a positive even multiple of the raster cell size."


class RasterPixelCountCheck(BaseRasterCheck):
    """Check if the grid does not contain more than a given amount of pixels, default 5 billion"""

    def __init__(self, *args, max_pixels=5e9, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_pixels = max_pixels

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            width, height = raster.shape
            return True if width * height <= self.max_pixels else False

    def description(self):
        return f"The file in {self.column_name} exceeds {self.max_pixels} pixels."


class RasterRangeCheck(BaseRasterCheck):
    """Check whether a raster has values outside of provided range.

    Also fails when there are no values in the raster.
    """

    def __init__(
        self,
        min_value=None,
        max_value=None,
        left_inclusive=True,
        right_inclusive=True,
        message=None,
        *args,
        **kwargs,
    ):
        if min_value is None and max_value is None:
            raise ValueError("Please supply at least one of {min_value, max_value}.")
        self.min_value = min_value
        self.max_value = max_value
        self.left_inclusive = left_inclusive
        self.right_inclusive = right_inclusive
        self.message = message
        super().__init__(*args, **kwargs)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            try:
                raster_min, raster_max = raster.min_max
            except RasterInterface.NoData:
                return False  # no data in the raster is invalid too

        if raster_min is None or raster_max is None:
            return False
        if self.min_value is not None:
            if self.left_inclusive and raster_min < self.min_value:
                return False
            if not self.left_inclusive and raster_min <= self.min_value:
                return False
        if self.max_value is not None:
            if self.right_inclusive and raster_max > self.max_value:
                return False
            if not self.right_inclusive and raster_max >= self.max_value:
                return False

        return True

    def description(self):
        if self.message:
            return self.message
        parts = []
        if self.min_value is not None:
            parts.append(f"{'<' if self.left_inclusive else '<='}{self.min_value}")
        if self.max_value is not None:
            parts.append(f"{'>' if self.right_inclusive else '>='}{self.max_value}")
        return f"{self.column_name} has values {' and/or '.join(parts)} or is empty"


class RasterCompressionUsedCheck(BaseRasterCheck):
    """Checks whether compression was used for the raster"""

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            return raster.compression != "NONE"

    def description(self):
        return f"Raster {self.column_name} is not compressed. It is recommended to use DEFLATE compression. This speeds up uploading and downloading and reduces storage space."


@dataclass
class GDALUnavailable:
    id: int


class GDALAvailableCheck(BaseRasterCheck):
    """Checks whether GDAL is available"""

    def get_invalid(self, session):
        context = session.model_checker_context
        raster_interface = context.raster_interface
        if not raster_interface.available():
            return [GDALUnavailable(id=1)]
        else:
            return []

    def description(self) -> str:
        return "raster checks require GDAL to be available"
